#ifndef ALLEGRO_PRIMITIVES_H
#define ALLEGRO_PRIMITIVES_H

#if (defined ALLEGRO_MINGW32) || (defined ALLEGRO_MSVC) || (defined ALLEGRO_BCC32)
   #ifndef ALLEGRO_STATICLINK
      #ifdef A5_PRIM_SRC
         #define _A5_PRIM_DLL __declspec(dllexport)
      #else
         #define _A5_PRIM_DLL __declspec(dllimport)
      #endif
   #else
      #define _A5_PRIM_DLL
   #endif
#endif

#if defined ALLEGRO_MSVC
   #define A5_PRIM_FUNC(type, name, args)      _A5_PRIM_DLL type __cdecl name args
#elif defined ALLEGRO_MINGW32
   #define A5_PRIM_FUNC(type, name, args)      extern type name args
#elif defined ALLEGRO_BCC32
   #define A5_PRIM_FUNC(type, name, args)      extern _A5_PRIM_DLL type name args
#else
   #define A5_PRIM_FUNC      AL_FUNC
#endif

#ifndef AL_PRIM_INLINE
  #ifdef __cplusplus
	 #define AL_PRIM_INLINE(type, name, args, code)    \
		static inline type name args;             \
		static inline type name args code
  #else
	 #define AL_PRIM_INLINE(type, name, args, code)    \
		extern __inline__ type name args;         \
		extern __inline__ type name args code
  #endif
#endif

#ifdef __cplusplus
extern "C" {
#endif

#include <allegro5/allegro5.h>

/*
 * Section: Primitives Addon Includes
 */

/* Enum: ALLEGRO_PRIM_TYPE
 *
 * Enumerates the types of primitves this addon can draw.
 *
 * ALLEGRO_PRIM_LINE_LIST - A list of lines, sequential pairs of vertices define disjointed lines
 *
 * ALLEGRO_PRIM_LINE_STRIP - A strip of lines, sequential vertices define a strip of lines
 *
 * ALLEGRO_PRIM_LINE_LOOP - Like a line strip, except at the end the first and the last vertices are also connected
 */
enum ALLEGRO_PRIM_TYPE {
   ALLEGRO_PRIM_LINE_LIST,
   ALLEGRO_PRIM_LINE_STRIP,
   ALLEGRO_PRIM_LINE_LOOP,
   ALLEGRO_PRIM_NUM_TYPES
};

/* Enum: ALLEGRO_VBUFFER_FLAGS
 *
 * Enumerates the flags that describe vertex buffers, and access to them.
 *
 * ALLEGRO_VBUFFER_SOFT - A software buffer, can be drawn by any backend
 *
 * ALLEGRO_VBUFFER_VIDEO - A video buffer, can be drawn by the software backend,
 *                         and the hardware accelerated backend that created it
 *
 * ALLEGRO_VBUFFER_LOCKED - Signifies that the vertex buffer is locked, meaning you can access its values. A buffer
 *                          has to be unlocked before it can be drawn
 *
 * ALLEGRO_VBUFFER_READ - Signifies that a buffer is read only, or that you are requesting read access
 *
 * ALLEGRO_VBUFFER_WRITE - Signifies that a buffer is write only, or that you are requesting read access
 */
enum ALLEGRO_VBUFFER_FLAGS {
   ALLEGRO_VBUFFER_SOFT =   1 << 0,
   ALLEGRO_VBUFFER_VIDEO =  1 << 1,
   ALLEGRO_VBUFFER_LOCKED = 1 << 2,
   ALLEGRO_VBUFFER_READ =   1 << 3,
   ALLEGRO_VBUFFER_WRITE =  1 << 4
};

/* Enum: ALLEGRO_PRIM_FLAGS
 *
 * Enumerates the flags that control the global operation of the entire addon. All are off by default.
 *
 * ALLEGRO_PRIM_LIGHT - Signifies whether the lighting engine should be used
 *
 * ALLEGRO_PRIM_3D - Signifies whether the addon should operate in 2D mode or 3D mode.
 *
 * ALLEGRO_PRIM_ZBUFFER - Signifies whether the software Z-buffer should be used
 *
 * ALLEGRO_PRIM_NORMAL - Signifies whether the normals of vertices should be used
 */
enum ALLEGRO_PRIM_FLAGS {
   ALLEGRO_PRIM_LIGHT,
   ALLEGRO_PRIM_3D,
   ALLEGRO_PRIM_ZBUFFER,
   ALLEGRO_PRIM_NORMAL,
   ALLEGRO_PRIM_NUM_FLAGS
};

/* Constant: ALLEGRO_VERTEX_CACHE_SIZE
 *
 * Defines the size of the transformation vertex cache for the software renderer. If you pass less than this many
 * vertices to the primitve rendering functions you will get a speed boost.
 * >#define ALLEGRO_VERTEX_CACHE_SIZE 256
 */
#define ALLEGRO_VERTEX_CACHE_SIZE 256

/* Constant: ALLEGRO_VBUFF_CACHE_SIZE
 *
 * Defines the size of the cache vertex buffer, used for the high-level primitives.
 * This corresponds to the maximum number of line segments that will be used to form them.
 * >#define ALLEGRO_VBUFF_CACHE_SIZE 256
 */
#define ALLEGRO_VBUFF_CACHE_SIZE 256

/* Constant: ALLEGRO_PRIM_QUALITY
 *
 * Defines the quality of the quadratic primitives. At 10, this roughly corresponds to error < 0.5 pixel.
 * >#define ALLEGRO_PRIM_QUALITY 10
 */
#define ALLEGRO_PRIM_QUALITY 10

/* Struct: ALLEGRO_VBUFFER
 *
 * Defines the generic vertex buffer type. They hold vertices in an arrangement that is conducive to rapid
 * drawing. Vertex buffers have to be locked to read/write data from them, and have to be unlocked while drawing.
 * Use the various access routines to operate on them, never touching the internal data if possible.
 */
typedef struct ALLEGRO_VBUFFER {
   void* data;
   void* priv;
   int lock_start;
   int lock_end;
   int len;
   int flags;
} ALLEGRO_VBUFFER;

/* Struct: ALLEGRO_VERTEX
 *
 * Defines the generic vertex type, with a 3D position, 3D normal, 4 component colour and texture coordinates
 * for a single texture.
 */
typedef struct {
   /* Variable: x, y, z
    *
    * 3D position
    */
   float x, y, z;
   /* Variable: nx, ny, nz
    *
    * 3D normal
    */
   float nx, ny, nz;
   /* Variable: r, g, b, a
    *
    * Colour
    */
   float r, g, b, a;
   /* Variable: u, v
    *
    * Texture position
    */
   float u, v;
} ALLEGRO_VERTEX;

/* Struct: ALLEGRO_TRANSFORM
 *
 * Defines the generic transformation type, a 4x4 matrix. 2D transforms use only a small subsection of this matrix,
 * namely the top left 2x2 matrix, and the right most 2x1 matrix, for a total of 6 values.
 */
typedef float ALLEGRO_TRANSFORM[4][4];

#include "inline/primitives.inl"

/*
 * Primary Functions
 */
A5_PRIM_FUNC(int, al_prim_draw, (ALLEGRO_VBUFFER* vbuff, ALLEGRO_BITMAP* texture, int start, int end, int type));
A5_PRIM_FUNC(int, al_prim_draw_indexed, (ALLEGRO_VBUFFER* vbuff, ALLEGRO_BITMAP* texture, int* indices, int num_vtx, int type));

/*
 * VBuff stuff
 */
A5_PRIM_FUNC(ALLEGRO_VBUFFER*, al_create_vbuff, (int len, int type));
A5_PRIM_FUNC(void, al_destroy_vbuff, (ALLEGRO_VBUFFER* vbuff));

A5_PRIM_FUNC(void, al_prim_unlock_vbuff, (ALLEGRO_VBUFFER* vbuff));
A5_PRIM_FUNC(int, al_prim_lock_vbuff, (ALLEGRO_VBUFFER* vbuff, int type));
A5_PRIM_FUNC(int, al_prim_lock_vbuff_range, (ALLEGRO_VBUFFER* vbuff, int start, int end, int type));
A5_PRIM_FUNC(int, al_prim_is_vbuff_locked, (ALLEGRO_VBUFFER* vbuff));
A5_PRIM_FUNC(int, al_prim_is_vbuff_range_locked, (ALLEGRO_VBUFFER* vbuff, int start, int end));
A5_PRIM_FUNC(void, al_prim_set_pos, (ALLEGRO_VBUFFER* vbuff, int idx, float x, float y, float z));
A5_PRIM_FUNC(void, al_prim_set_normal, (ALLEGRO_VBUFFER* vbuff, int idx, float nx, float ny, float nz));
A5_PRIM_FUNC(void, al_prim_set_uv, (ALLEGRO_VBUFFER* vbuff, int idx, float u, float v));
A5_PRIM_FUNC(void, al_prim_set_color, (ALLEGRO_VBUFFER* vbuff, int idx, const ALLEGRO_COLOR col));

A5_PRIM_FUNC(void, al_prim_set_vertex, (ALLEGRO_VBUFFER* vbuff, int idx, const ALLEGRO_VERTEX *vtx));
A5_PRIM_FUNC(void, al_prim_get_vertex, (ALLEGRO_VBUFFER* vbuff, int idx, ALLEGRO_VERTEX *vtx));

/*
 * Other
 */
A5_PRIM_FUNC(int, al_is_bitmap_region_locked, (ALLEGRO_BITMAP* bmp, int x1, int y1, int x2, int y2));

/*
 * Transformations
 */
A5_PRIM_FUNC(void, al_trans_use_trans, (ALLEGRO_TRANSFORM* trans));
A5_PRIM_FUNC(void, al_trans_copy, (ALLEGRO_TRANSFORM* src, ALLEGRO_TRANSFORM* dest));

/*
 *High level primitives
 */
A5_PRIM_FUNC(void, al_draw_line_2d, (float x1, float y1, float x2, float y2, ALLEGRO_COLOR color1, ALLEGRO_COLOR color2, ALLEGRO_BITMAP* texture));
A5_PRIM_FUNC(void, al_draw_triangle_2d, (float x1, float y1, float x2, float y2, float x3, float y3, ALLEGRO_COLOR color, ALLEGRO_BITMAP* texture, int type));
A5_PRIM_FUNC(void, al_draw_rectangle_2d, (float x1, float y1, float x2, float y2, ALLEGRO_COLOR color, ALLEGRO_BITMAP* texture, int type));

A5_PRIM_FUNC(void, al_calc_arc_2d, (ALLEGRO_VBUFFER* vbuff, float cx, float cy, float rx, float ry, float theta, float delta_theta, int start, int num_segments));
A5_PRIM_FUNC(void, al_draw_circle_2d, (float cx, float cy, float r, ALLEGRO_COLOR color1, ALLEGRO_COLOR color2, ALLEGRO_BITMAP* texture, int type));
A5_PRIM_FUNC(void, al_draw_ellipse_2d, (float cx, float cy, float rx, float ry, ALLEGRO_COLOR color1, ALLEGRO_COLOR color2, ALLEGRO_BITMAP* texture, int type));
A5_PRIM_FUNC(void, al_draw_arc_2d, (float cx, float cy, float r, float start_theta, float delta_theta, ALLEGRO_COLOR color1, ALLEGRO_COLOR color2, ALLEGRO_BITMAP* texture, int type));

A5_PRIM_FUNC(void, al_calc_spline_2d, (ALLEGRO_VBUFFER* vbuff, float points[8], int start, int num_points));
A5_PRIM_FUNC(void, al_draw_spline_2d, (float points[8], ALLEGRO_COLOR color1, ALLEGRO_COLOR color2, ALLEGRO_BITMAP* texture));


#ifdef __cplusplus
}
#endif

#endif
