/*         ______   ___    ___
 *        /\  _  \ /\_ \  /\_ \
 *        \ \ \L\ \\//\ \ \//\ \      __     __   _ __   ___
 *         \ \  __ \ \ \ \  \ \ \   /'__`\ /'_ `\/\`'__\/ __`\
 *          \ \ \/\ \ \_\ \_ \_\ \_/\  __//\ \L\ \ \ \//\ \L\ \
 *           \ \_\ \_\/\____\/\____\ \____\ \____ \ \_\\ \____/
 *            \/_/\/_/\/____/\/____/\/____/\/___L\ \/_/ \/___/
 *                                           /\____/
 *                                           \_/__/
 *
 *      Some high level routines, provided for user's convinience.
 *
 *
 *      By Pavel Sountsov.
 *
 *
 *      Bezier spline plotter By Seymour Shlien.
 *
 *      Optimised version by Sven Sandberg.
 *
 *      I'm not sure wether or not we still use the Castelau Algorithm
 *      described in the book :o)
 *
 *      Interactive Computer Graphics
 *      by Peter Burger and Duncan Gillies
 *      Addison-Wesley Publishing Co 1989
 *      ISBN 0-201-17439-1
 *
 *      The 4 th order Bezier curve is a cubic curve passing
 *      through the first and fourth point. The curve does
 *      not pass through the middle two points. They are merely
 *      guide points which control the shape of the curve. The
 *      curve is tangent to the lines joining points 1 and 2
 *      and points 3 and 4.
 *
 *      See readme.txt for copyright information.
 */


#include "allegro5/a5_primitives.h"

ALLEGRO_VBUFFER* cache_buffer;

static void verify_cache()
{
   if (!cache_buffer)
      cache_buffer = al_create_vbuff(ALLEGRO_VBUFF_CACHE_SIZE, ALLEGRO_VBUFFER_SOFT | ALLEGRO_VBUFFER_WRITE | ALLEGRO_VBUFFER_READ);
}

/*
 * Section: High level primitives
 */

/* Function: al_draw_line_2d
 *
 * Draws a line segment between two points, blending between two colours and using a provided texture.
 *
 * Parameters:
 *
 * x1, y1, x2, y2 - Start and end points of the line
 * color1, color2 - Colors to blend between
 * texture - Texture to use, pass 0 to use no texture
 */
void al_draw_line_2d(float x1, float y1, float x2, float y2, ALLEGRO_COLOR color1, ALLEGRO_COLOR color2, ALLEGRO_BITMAP* texture)
{
   verify_cache();
   al_prim_lock_vbuff_range(cache_buffer, 0, 2, ALLEGRO_VBUFFER_WRITE);
   
   al_prim_set_pos(cache_buffer, 0, x1, y1, 0);
   al_prim_set_pos(cache_buffer, 1, x2, y2, 0);
   
   al_prim_set_color(cache_buffer, 0, color1);
   al_prim_set_color(cache_buffer, 1, color2);
   
   if (texture) {
      al_prim_set_uv(cache_buffer, 0, 0, 0);
      al_prim_set_uv(cache_buffer, 1, 1, 1);
   }
   
   al_prim_unlock_vbuff(cache_buffer);
   
   al_prim_draw(cache_buffer, texture, 0, 2, ALLEGRO_PRIM_LINE_LIST);
}

/* Function: al_draw_triangle_2d
 *
 * Draws a triangle.
 *
 * Parameters:
 *
 * x1, y1, x2, y2, x3, y3 - Three points of the triangle
 * color - Color to use
 * texture - Texture to use, pass 0 to use no texture
 * type - ALLEGRO_FILLED or ALLEGRO_OUTLINED
 */
void al_draw_triangle_2d(float x1, float y1, float x2, float y2, float x3, float y3, ALLEGRO_COLOR color, ALLEGRO_BITMAP* texture, int type)
{
   verify_cache();
   if (type == ALLEGRO_OUTLINED) {
      al_prim_lock_vbuff_range(cache_buffer, 0, 3, ALLEGRO_VBUFFER_WRITE);
      
      al_prim_set_pos(cache_buffer, 0, x1, y1, 0);
      al_prim_set_pos(cache_buffer, 1, x2, y2, 0);
      al_prim_set_pos(cache_buffer, 2, x3, y3, 0);
      
      al_prim_set_color(cache_buffer, 0, color);
      al_prim_set_color(cache_buffer, 1, color);
      al_prim_set_color(cache_buffer, 2, color);
      
      if (texture) {
         al_prim_set_uv(cache_buffer, 0, 0, 0);
         al_prim_set_uv(cache_buffer, 1, 0, 1);
         al_prim_set_uv(cache_buffer, 2, 1, 1);
      }
      
      al_prim_unlock_vbuff(cache_buffer);
      
      al_prim_draw(cache_buffer, texture, 0, 3, ALLEGRO_PRIM_LINE_LOOP);
   }
}

/* Function: al_draw_rectangle_2d
 *
 * Draws a rectangle.
 *
 * Parameters:
 *
 * x1, y1, x2, y2 - Upper left and lower right points of a triangle
 * color - Color to use
 * texture - Texture to use, pass 0 to use no texture
 * type - ALLEGRO_FILLED or ALLEGRO_OUTLINED
 */
void al_draw_rectangle_2d(float x1, float y1, float x2, float y2, ALLEGRO_COLOR color, ALLEGRO_BITMAP* texture, int type)
{
   verify_cache();
   if (type == ALLEGRO_OUTLINED) {
      al_prim_lock_vbuff_range(cache_buffer, 0, 4, ALLEGRO_VBUFFER_WRITE);
      
      al_prim_set_pos(cache_buffer, 0, x1, y1, 0);
      al_prim_set_pos(cache_buffer, 1, x2, y1, 0);
      al_prim_set_pos(cache_buffer, 2, x2, y2, 0);
      al_prim_set_pos(cache_buffer, 3, x1, y2, 0);
      
      al_prim_set_color(cache_buffer, 0, color);
      al_prim_set_color(cache_buffer, 1, color);
      al_prim_set_color(cache_buffer, 2, color);
      al_prim_set_color(cache_buffer, 3, color);
      
      if (texture) {
         al_prim_set_uv(cache_buffer, 0, 0, 0);
         al_prim_set_uv(cache_buffer, 1, 0, 1);
         al_prim_set_uv(cache_buffer, 2, 1, 1);
         al_prim_set_uv(cache_buffer, 3, 1, 0);
      }
      
      al_prim_unlock_vbuff(cache_buffer);
      
      al_prim_draw(cache_buffer, texture, 0, 4, ALLEGRO_PRIM_LINE_LOOP);
   }
}

/* Function: al_calc_arc_2d
 *
 * Calculates an elliptical arc, and sets the vertices in a buffer to the calculated positions
 *
 * Parameters:
 *
 * vbuff - Vertex buffer to fill out
 * cx, cy - Center of the arc
 * rx, ry - Radii of the arc
 * start_theta - The initial angle from which the arc is calculated
 * delta_theta - Angular span of the arc (pass a negative number to switch direction)
 * start - The index of the first calculated point in the vertex buffer
 * num_points - The number of points to calculate
 */
void al_calc_arc_2d(ALLEGRO_VBUFFER* vbuff, float cx, float cy, float rx, float ry, float start_theta, float delta_theta, int start, int num_points)
{
   ASSERT(vbuff);
   
   int need_unlock = 0;
   float theta;
   float tangetial_factor;
   float radial_factor;
   float x;
   float y;
   int ii;
   
   if (al_prim_is_vbuff_locked(vbuff)) {
      if (!al_prim_is_vbuff_range_locked(vbuff, start, start + num_points))
         return;
   } else {
      if (!al_prim_lock_vbuff_range(cache_buffer, start, start + num_points, ALLEGRO_VBUFFER_WRITE))
         return;
      need_unlock = 1;
   }
   
   theta = delta_theta / ((float)num_points - 1);
   tangetial_factor = tanf(theta);
   radial_factor = cosf(theta);
   x = cosf(start_theta);
   y = sinf(start_theta);
   
   for (ii = 0; ii < num_points; ii++) {
      al_prim_set_pos(vbuff, ii + start, rx * x + cx, ry * y + cy, 0);
      
      float tx = -y;
      float ty = x;
      
      x += tx * tangetial_factor;
      y += ty * tangetial_factor;
      
      x *= radial_factor;
      y *= radial_factor;
   }
   
   if (need_unlock)
      al_prim_unlock_vbuff(vbuff);
}

/* Function: al_draw_ellipse_2d
 *
 * Draws an ellipse, blending between two colours.
 *
 * Parameters:
 *
 * cx, cy - Center of the ellipse
 * rx, ry - Radii of the ellipse
 * color1 - Color of the center of the ellipse
 * color2 - Color of the edge of the ellipse
 * texture - Texture to use, pass 0 to use no texture
 * type - ALLEGRO_FILLED or ALLEGRO_OUTLINED
 */
void al_draw_ellipse_2d(float cx, float cy, float rx, float ry, ALLEGRO_COLOR color1, ALLEGRO_COLOR color2, ALLEGRO_BITMAP* texture, int type)
{
   verify_cache();
   if (type == ALLEGRO_OUTLINED) {
      int num_segments = ALLEGRO_PRIM_QUALITY * sqrtf((rx + ry) / 2.0f);
      int ii;
      float du;
      float u;
      
      if (num_segments >= ALLEGRO_VBUFF_CACHE_SIZE) {
         num_segments = ALLEGRO_VBUFF_CACHE_SIZE - 1;
      }
      
      al_calc_arc_2d(cache_buffer, cx, cy, rx, ry, 0, AL_PI * 2, 0, num_segments);
      
      du = 1 / (float)(num_segments - 1);
      u = 0;
      
      for (ii = 0; ii < num_segments - 1; ii++) {
         al_prim_set_color(cache_buffer, ii, color2);
         if (texture) {
            al_prim_set_uv(cache_buffer, ii, u, 0);
            u += du;
         }
      }
      
      al_prim_draw(cache_buffer, texture, 0, num_segments - 1, ALLEGRO_PRIM_LINE_LOOP);
   }
}

/* Function: al_draw_circle_2d
 *
 * Draws a circle, blending between two colours.
 *
 * Parameters:
 *
 * cx, cy - Center of the circle
 * r - Radius of the circle
 * color1 - Color of the center of the circle
 * color2 - Color of the edge of the circle
 * texture - Texture to use, pass 0 to use no texture
 * type - ALLEGRO_FILLED or ALLEGRO_OUTLINED
 */
void al_draw_circle_2d(float cx, float cy, float r, ALLEGRO_COLOR color1, ALLEGRO_COLOR color2, ALLEGRO_BITMAP* texture, int type)
{
   al_draw_ellipse_2d(cx, cy, r, r, color1, color2, texture, type);
}

/* Function: al_draw_arc_2d
 *
 * Draws an arc, blending between two colours.
 *
 * Parameters:
 *
 * cx, cy - Center of the arc
 * r - Radius of the arc
 * color1 - Color of the start of the arc
 * color2 - Color of the end of the arc
 * start_theta - The initial angle from which the arc is calculated
 * delta_theta - Angular span of the arc (pass a negative number to switch direction)
 * texture - Texture to use, pass 0 to use no texture
 * type - ALLEGRO_FILLED or ALLEGRO_OUTLINED. A filled arc is a sector.
 */
void al_draw_arc_2d(float cx, float cy, float r, float start_theta, float delta_theta, ALLEGRO_COLOR color1, ALLEGRO_COLOR color2, ALLEGRO_BITMAP* texture, int type)
{
   verify_cache();
   if (type == ALLEGRO_OUTLINED) {
      int num_segments = delta_theta / (2 * AL_PI) * ALLEGRO_PRIM_QUALITY * sqrtf(r);
      float du;
      float u;
      int ii;
      ALLEGRO_COLOR cur_color, delta_color;
      
      if (num_segments >= ALLEGRO_VBUFF_CACHE_SIZE) {
         num_segments = ALLEGRO_VBUFF_CACHE_SIZE - 1;
      }
      
      al_calc_arc_2d(cache_buffer, cx, cy, r, r, start_theta, delta_theta, 0, num_segments);
      
      du = 1 / (float)num_segments;
      u = 0;
      cur_color = color1;
      delta_color.r = (color2.r - color1.r) / (float)num_segments;
      delta_color.g = (color2.g - color1.g) / (float)num_segments;
      delta_color.b = (color2.b - color1.b) / (float)num_segments;
      delta_color.a = (color2.a - color1.a) / (float)num_segments;
      
      for (ii = 0; ii < num_segments; ii++) {
         al_prim_set_color(cache_buffer, ii, cur_color);
         
         cur_color.r += delta_color.r;
         cur_color.g += delta_color.g;
         cur_color.b += delta_color.b;
         cur_color.a += delta_color.a;
         
         if (texture) {
            al_prim_set_uv(cache_buffer, ii, u, 0);
            u += du;
         }
      }
      
      al_prim_draw(cache_buffer, texture, 0, num_segments, ALLEGRO_PRIM_LINE_STRIP);
   }
}

/* Function: al_calc_spline_2d
 *
 * Calculates a spline given 4 control points.
 *
 * Parameters:
 *
 * vbuff - Vertex buffer to fill out
 * points - An array of 4 pairs of coordinates of the 4 control points
 * start - The index of the first calculated point in the vertex buffer
 * num_points - The number of points to calculate
 */
void al_calc_spline_2d(ALLEGRO_VBUFFER* vbuff, float points[8], int start, int num_points)
{
   ASSERT(vbuff);
   
   int need_unlock = 0;
   /* Derivatives of x(t) and y(t). */
   float x, dx, ddx, dddx;
   float y, dy, ddy, dddy;
   int ii;
   
   /* Temp variables used in the setup. */
   float dt, dt2, dt3;
   float xdt2_term, xdt3_term;
   float ydt2_term, ydt3_term;
   
   if (al_prim_is_vbuff_locked(vbuff)) {
      if (!al_prim_is_vbuff_range_locked(vbuff, start, start + num_points))
         return;
   } else {
      if (!al_prim_lock_vbuff_range(cache_buffer, start, start + num_points, ALLEGRO_VBUFFER_WRITE))
         return;
      need_unlock = 1;
   }
   
   dt = 1.0 / (num_points - 1);
   dt2 = (dt * dt);
   dt3 = (dt2 * dt);
   
   /* x coordinates. */
   xdt2_term = 3 * (points[4] - 2 * points[2] + points[0]);
   xdt3_term = points[6] + 3 * (-points[4] + points[2]) - points[0];
   
   xdt2_term = dt2 * xdt2_term;
   xdt3_term = dt3 * xdt3_term;
   
   dddx = 6 * xdt3_term;
   ddx = -6 * xdt3_term + 2 * xdt2_term;
   dx = xdt3_term - xdt2_term + 3 * dt * (points[2] - points[0]);
   x = points[0];
   
   /* y coordinates. */
   ydt2_term = 3 * (points[5] - 2 * points[3] + points[1]);
   ydt3_term = points[7] + 3 * (-points[5] + points[3]) - points[1];
   
   ydt2_term = dt2 * ydt2_term;
   ydt3_term = dt3 * ydt3_term;
   
   dddy = 6 * ydt3_term;
   ddy = -6 * ydt3_term + 2 * ydt2_term;
   dy = ydt3_term - ydt2_term + dt * 3 * (points[3] - points[1]);
   y = points[1];
   
   al_prim_set_pos(vbuff, start, x, y, 0);
   
   for (ii = 1; ii < num_points; ii++) {
      ddx += dddx;
      dx += ddx;
      x += dx;
      
      ddy += dddy;
      dy += ddy;
      y += dy;
      
      al_prim_set_pos(vbuff, start + ii, x, y, 0);
   }
   
   if (need_unlock)
      al_prim_unlock_vbuff(vbuff);
}

/* Function: al_draw_spline_2d
 *
 * Draws a spline given 4 control points, blending between two colours.
 *
 * Parameters:
 *
 * points - An array of 4 pairs of coordinates of the 4 control points
 * color2 - Color of the start of the spline
 * color2 - Color of the end of the spline
 * texture - Texture to use, pass 0 to use no texture
 */
void al_draw_spline_2d(float points[8], ALLEGRO_COLOR color1, ALLEGRO_COLOR color2, ALLEGRO_BITMAP* texture)
{
   float du;
   float u;
   ALLEGRO_COLOR cur_color;
   ALLEGRO_COLOR delta_color;
   int ii;
   int num_points = (int)(sqrtf(hypotf(points[2] - points[0], points[3] - points[1]) +
                                hypotf(points[4] - points[2], points[5] - points[3]) +
                                hypotf(points[6] - points[4], points[7] - points[5])) *
                          1.2 * ALLEGRO_PRIM_QUALITY / 10);
                          
   verify_cache();
   
   if (num_points >= ALLEGRO_VBUFF_CACHE_SIZE) {
      num_points = ALLEGRO_VBUFF_CACHE_SIZE - 1;
   }
   
   al_calc_spline_2d(cache_buffer, points, 0, num_points);
   
   du = 1 / (float)num_points;
   u = 0;
   cur_color = color1;
   delta_color.r = (color2.r - color1.r) / (float)num_points;
   delta_color.g = (color2.g - color1.g) / (float)num_points;
   delta_color.b = (color2.b - color1.b) / (float)num_points;
   delta_color.a = (color2.a - color1.a) / (float)num_points;
   
   for (ii = 0; ii < num_points; ii++) {
      al_prim_set_color(cache_buffer, ii, cur_color);
      
      cur_color.r += delta_color.r;
      cur_color.g += delta_color.g;
      cur_color.b += delta_color.b;
      cur_color.a += delta_color.a;
      
      if (texture) {
         al_prim_set_uv(cache_buffer, ii, u, 0);
         u += du;
      }
   }
   
   al_prim_draw(cache_buffer, texture, 0, num_points, ALLEGRO_PRIM_LINE_STRIP);
}
