/*         ______   ___    ___
 *        /\  _  \ /\_ \  /\_ \
 *        \ \ \L\ \\//\ \ \//\ \      __     __   _ __   ___
 *         \ \  __ \ \ \ \  \ \ \   /'__`\ /'_ `\/\`'__\/ __`\
 *          \ \ \/\ \ \_\ \_ \_\ \_/\  __//\ \L\ \ \ \//\ \L\ \
 *           \ \_\ \_\/\____\/\____\ \____\ \____ \ \_\\ \____/
 *            \/_/\/_/\/____/\/____/\/____/\/___L\ \/_/ \/___/
 *                                           /\____/
 *                                           \_/__/
 *
 *      Software implementation of some of the primitive routines.
 *
 *
 *      By Pavel Sountsov.
 *
 *      See readme.txt for copyright information.
 */


#include "allegro5/allegro5.h"
#include "allegro5/internal/aintern.h"
#include ALLEGRO_INTERNAL_HEADER
#include "allegro5/internal/aintern_system.h"
#include "allegro5/internal/aintern_display.h"
#include "allegro5/internal/aintern_bitmap.h"

#include "allegro5/a5_primitives.h"
#include "allegro5/internal/aintern_prim_soft.h"

extern ALLEGRO_TRANSFORM _al_global_trans;

/*
The vertex cache allows for bulk transformation of vertices, for faster run speeds
*/
ALLEGRO_VERTEX vertex_cache[ALLEGRO_VERTEX_CACHE_SIZE];

void _al_create_vbuff_soft(ALLEGRO_VBUFFER* vbuff)
{
   vbuff->data = (ALLEGRO_VERTEX*)malloc(sizeof(ALLEGRO_VERTEX) * vbuff->len);
   memset(vbuff->data, sizeof(ALLEGRO_VERTEX) * vbuff->len, 0);
}

void _al_destroy_vbuff_soft(ALLEGRO_VBUFFER* vbuff)
{
   free(vbuff->data);
}

int _al_prim_draw_soft(ALLEGRO_BITMAP* texture, ALLEGRO_VBUFFER* vbuff, int start, int end, int type)
{
   ASSERT(!al_prim_is_vbuff_locked(vbuff));
   
   int num_primitives = 0;
   int num_vtx = end - start;
   int use_cache = num_vtx < ALLEGRO_VERTEX_CACHE_SIZE;
   
   if (!al_prim_lock_vbuff_range(vbuff, start, end, ALLEGRO_VBUFFER_READ))
      return 0;
      
   if (use_cache) {
      int ii;
      int n = 0;
      for (ii = start; ii < end; ii++) {
         al_prim_get_vertex(vbuff, ii, &vertex_cache[n]);
         
         al_trans_trans_vertex_2d(&_al_global_trans, &vertex_cache[n]);
         //printf("X: %f\n", vertex_cache[n].x);
         n++;
      }
   }
   
#define SET_VERTICES(idx1, idx2)                      \
   ALLEGRO_VERTEX v1, v2;                             \
   al_prim_get_vertex(vbuff, idx1, &v1);              \
   al_prim_get_vertex(vbuff, idx2, &v2);              \
   al_trans_trans_vertex_2d(&_al_global_trans, &v1);  \
   al_trans_trans_vertex_2d(&_al_global_trans, &v2);
   
   switch (type) {
   case ALLEGRO_PRIM_LINE_LIST: {
         if (use_cache) {
            int ii;
            for (ii = 0; ii < num_vtx - 1; ii += 2) {
               _al_line_2d(texture, vbuff, &vertex_cache[ii], &vertex_cache[ii + 1]);
            }
         } else {
            int ii;
            for (ii = start; ii < end - 1; ii += 2) {
               SET_VERTICES(ii, ii + 1)
               
               _al_line_2d(texture, vbuff, &v1, &v2);
            }
         }
         num_primitives = num_vtx / 2;
         break;
      };
   case ALLEGRO_PRIM_LINE_STRIP: {
         if (use_cache) {
            int ii;
            for (ii = 1; ii < num_vtx; ii++) {
               _al_line_2d(texture, vbuff, &vertex_cache[ii - 1], &vertex_cache[ii]);
            }
         } else {
            int ii;
            for (ii = start + 1; ii < end; ii++) {
               SET_VERTICES(ii - 1, ii)
               
               _al_line_2d(texture, vbuff, &v1, &v2);
            }
         }
         num_primitives = num_vtx - 1;
         break;
      };
   case ALLEGRO_PRIM_LINE_LOOP: {
         if (use_cache) {
            int ii;
            for (ii = 1; ii < num_vtx; ii++) {
               _al_line_2d(texture, vbuff, &vertex_cache[ii - 1], &vertex_cache[ii]);
            }
            _al_line_2d(texture, vbuff, &vertex_cache[0], &vertex_cache[num_vtx - 1]);
         } else {
            int ii;
            for (ii = start + 1; ii < end; ii++) {
               SET_VERTICES(ii - 1, ii)
               
               _al_line_2d(texture, vbuff, &v1, &v2);
            }
            SET_VERTICES(end - 1, start)
            _al_line_2d(texture, vbuff, &v1, &v2);
         }
         num_primitives = num_vtx;
         break;
      };
   }
   
   al_prim_unlock_vbuff(vbuff);
   
   return num_primitives;
}

int _al_prim_draw_indexed_soft(ALLEGRO_BITMAP* texture, ALLEGRO_VBUFFER* vbuff, int* indices, int num_vtx, int type)
{
   ASSERT(!al_prim_is_vbuff_locked(vbuff));
   
   int num_primitives = 0;
   
   int use_cache = 1;
   int min_idx = indices[0], max_idx = indices[0];
   
   /*
   Determine the range we are dealing with
   */
   int ii;
   for (ii = 1; ii < num_vtx; ii++) {
      int idx = indices[ii];
      if (max_idx < idx)
         max_idx = idx;
      else if (min_idx > indices[ii])
         min_idx = idx;
   }
   if (max_idx - min_idx >= ALLEGRO_VERTEX_CACHE_SIZE) {
      use_cache = 0;
   }
   
   if (!al_prim_lock_vbuff_range(vbuff, min_idx, max_idx, ALLEGRO_VBUFFER_READ))
      return 0;
      
   if (use_cache) {
      int ii;
      for (ii = 0; ii < num_vtx; ii++) {
         int idx = indices[ii];
         al_prim_get_vertex(vbuff, idx, &vertex_cache[idx - min_idx]);
         al_trans_trans_vertex_2d(&_al_global_trans, &vertex_cache[idx - min_idx]);
      }
   }
   
#define SET_VERTICES(idx1, idx2)                      \
   ALLEGRO_VERTEX v1, v2;                             \
   al_prim_get_vertex(vbuff, idx1, &v1);              \
   al_prim_get_vertex(vbuff, idx2, &v2);              \
   al_trans_trans_vertex_2d(&_al_global_trans, &v1);  \
   al_trans_trans_vertex_2d(&_al_global_trans, &v2);
   
   switch (type) {
   case ALLEGRO_PRIM_LINE_LIST: {
         if (use_cache) {
            int ii;
            for (ii = 0; ii < num_vtx - 1; ii += 2) {
               int idx1 = indices[ii] - min_idx;
               int idx2 = indices[ii + 1] - min_idx;
               
               _al_line_2d(texture, vbuff, &vertex_cache[idx1], &vertex_cache[idx2]);
            }
         } else {
            int ii;
            for (ii = 0; ii < num_vtx - 1; ii += 2) {
               int idx1 = indices[ii];
               int idx2 = indices[ii + 1];
               
               SET_VERTICES(idx1, idx2)
               _al_line_2d(texture, vbuff, &v1, &v2);
            }
         }
         num_primitives = num_vtx / 2;
         break;
      };
   case ALLEGRO_PRIM_LINE_STRIP: {
         if (use_cache) {
            int ii;
            for (ii = 1; ii < num_vtx; ii++) {
               int idx1 = indices[ii - 1] - min_idx;
               int idx2 = indices[ii] - min_idx;
               
               _al_line_2d(texture, vbuff, &vertex_cache[idx1], &vertex_cache[idx2]);
            }
         } else {
            int ii;
            for (ii = 1; ii < num_vtx; ii++) {
               int idx1 = indices[ii - 1];
               int idx2 = indices[ii];
               SET_VERTICES(idx1, idx2)
               
               _al_line_2d(texture, vbuff, &v1, &v2);
            }
         }
         num_primitives = num_vtx - 1;
         break;
      };
   case ALLEGRO_PRIM_LINE_LOOP: {
         if (use_cache) {
            int ii;
            
            for (ii = 1; ii < num_vtx; ii++) {
               int idx1 = indices[ii - 1] - min_idx;
               int idx2 = indices[ii] - min_idx;
               
               _al_line_2d(texture, vbuff, &vertex_cache[idx1], &vertex_cache[idx2]);
            }
            int idx1 = indices[0] - min_idx;
            int idx2 = indices[num_vtx - 1] - min_idx;
            
            _al_line_2d(texture, vbuff, &vertex_cache[idx1], &vertex_cache[idx2]);
         } else {
            int ii;
            for (ii = 1; ii < num_vtx; ii++) {
               int idx1 = indices[ii - 1];
               int idx2 = indices[ii];
               SET_VERTICES(idx1, idx2)
               
               _al_line_2d(texture, vbuff, &v1, &v2);
            }
            
            SET_VERTICES(indices[num_vtx - 1], indices[0])
            _al_line_2d(texture, vbuff, &v1, &v2);
         }
         num_primitives = num_vtx;
         break;
      };
   }
   
   al_prim_unlock_vbuff(vbuff);
   
   return num_primitives;
}

void _al_prim_unlock_vbuff_soft(ALLEGRO_VBUFFER* vbuff)
{
   vbuff->flags &= ~ALLEGRO_VBUFFER_LOCKED;
}

int _al_prim_lock_vbuff_range_soft(ALLEGRO_VBUFFER* vbuff, int start, int end, int type)
{
   vbuff->flags |= ALLEGRO_VBUFFER_LOCKED;
   vbuff->lock_start = start;
   vbuff->lock_end = end;
   return 1;
}

void _al_prim_set_pos_soft(ALLEGRO_VBUFFER* vbuff, int idx, float x, float y, float z)
{
   ALLEGRO_VERTEX* vtx = &((ALLEGRO_VERTEX*)vbuff->data)[idx];
   vtx->x = x;
   vtx->y = y;
   vtx->z = z;
}

void _al_prim_set_normal_soft(ALLEGRO_VBUFFER* vbuff, int idx, float nx, float ny, float nz)
{
   ALLEGRO_VERTEX* vtx = &((ALLEGRO_VERTEX*)vbuff->data)[idx];
   vtx->nx = nx;
   vtx->ny = ny;
   vtx->nz = nz;
}

void _al_prim_set_uv_soft(ALLEGRO_VBUFFER* vbuff, int idx, float u, float v)
{
   ALLEGRO_VERTEX* vtx = &((ALLEGRO_VERTEX*)vbuff->data)[idx];
   vtx->u = u;
   vtx->v = v;
}

void _al_prim_set_color_soft(ALLEGRO_VBUFFER* vbuff, int idx, const ALLEGRO_COLOR col)
{
   ALLEGRO_VERTEX* vtx = &((ALLEGRO_VERTEX*)vbuff->data)[idx];
   vtx->r = col.r;
   vtx->g = col.g;
   vtx->b = col.b;
   vtx->a = col.a;
}

void _al_prim_get_vertex_soft(ALLEGRO_VBUFFER* vbuff, int idx, ALLEGRO_VERTEX *vtx)
{
   *vtx = ((ALLEGRO_VERTEX*)vbuff->data)[idx];
}

void _al_prim_set_vertex_soft(ALLEGRO_VBUFFER* vbuff, int idx, const ALLEGRO_VERTEX *vtx)
{
   ((ALLEGRO_VERTEX*)vbuff->data)[idx] = *vtx;
}
