/*         ______   ___    ___
 *        /\  _  \ /\_ \  /\_ \
 *        \ \ \L\ \\//\ \ \//\ \      __     __   _ __   ___
 *         \ \  __ \ \ \ \  \ \ \   /'__`\ /'_ `\/\`'__\/ __`\
 *          \ \ \/\ \ \_\ \_ \_\ \_/\  __//\ \L\ \ \ \//\ \L\ \
 *           \ \_\ \_\/\____\/\____\ \____\ \____ \ \_\\ \____/
 *            \/_/\/_/\/____/\/____/\/____/\/___L\ \/_/ \/___/
 *                                           /\____/
 *                                           \_/__/
 *
 *      Core primitive addon functions.
 *
 *
 *      By Pavel Sountsov.
 *
 *      See readme.txt for copyright information.
 */

#include "allegro5/allegro5.h"
#include "allegro5/a5_primitives.h"
#include "allegro5/internal/aintern_prim_opengl.h"
#include "allegro5/internal/aintern_prim_directx.h"
#include "allegro5/a5_opengl.h"
/*
TODO: Figure out a way to determine which one to include and such
#include "allegro5/a5_direct3d.h"
*/
#define ALLEGRO_DIRECT3D 8
#include "allegro5/internal/aintern_prim_soft.h"
#include "allegro5/internal/aintern_bitmap.h"

/*
 * Section: Primitives Addon Source
 */

int _al_prim_global_flags[ALLEGRO_PRIM_NUM_FLAGS];
ALLEGRO_TRANSFORM _al_global_trans = {
   {1, 0, 0, 0},
   {0, 1, 0, 0},
   {0, 0, 1, 0},
   {0, 0, 0, 1}
};

static void temp_trans(float x, float y)
{
   int flags = al_get_display_flags();
   al_trans_translate_2d(&_al_global_trans, x, y);
   if (flags & ALLEGRO_OPENGL) {
      _al_trans_use_trans_opengl(&_al_global_trans);
   } else if (flags & ALLEGRO_DIRECT3D) {
      _al_trans_use_trans_directx(&_al_global_trans);
   }
}

/*
 * Section: Core Drawing Routines
 */

/* Function: al_prim_draw
 *
 * Draws a subset of the passed vertex buffer.
 *
 * Parameters:
 *
 * texture - Texture to use, pass 0 to use only shaded primitves
 * vbuff - Vertex buffer to use
 * start, end - Start and end of the subset of the vertex buffer to draw
 * type - Primitive type to draw
 *
 * Returns:
 *
 * Number of primitives drawn
 *
 * See Also:
 *
 * <ALLEGRO_VBUFFER>, <ALLEGRO_PRIM_TYPE>, <al_prim_draw_indexed>
 */
int al_prim_draw(ALLEGRO_VBUFFER* vbuff, ALLEGRO_BITMAP* texture, int start, int end, int type)
{
   ASSERT(vbuff);
   ASSERT(end >= start);
   ASSERT(type > 0 && type < ALLEGRO_PRIM_NUM_TYPES);
   
   ALLEGRO_BITMAP *target = al_get_target_bitmap();
   int ret = 0;
   int flags = al_get_display_flags();
   
   /* In theory, if we ever get a camera concept for this addon, the transformation into
    * view space should occur here
    */
   
   if (target->parent) {
      /*
       * For sub-bitmaps
       */
      temp_trans((float)target->xofs, (float)target->yofs);
   }
   
   if (target->flags & ALLEGRO_MEMORY_BITMAP) {
      ret =  _al_prim_draw_soft(texture, vbuff, start, end, type);
   } else {
      if (flags & ALLEGRO_OPENGL) {
         ret =  _al_prim_draw_opengl(texture, vbuff, start, end, type);
      } else if (flags & ALLEGRO_DIRECT3D) {
         ret =  _al_prim_draw_directx(texture, vbuff, start, end, type);
      }
   }
   
   if (target->parent) {
      /*
       * Move it back, if rounding errors become an issue we can do a copy-restore instead
       */
      temp_trans(-(float)target->xofs, -(float)target->yofs);
   }
   
   return ret;
}

/* Function: al_prim_draw_indexed
 *
 * Draws a subset of the passed vertex buffer. This function uses an index array to specify which vertices to use.
 *
 * Parameters:
 *
 * texture - Texture to use, pass 0 to use only shaded primitves
 * vbuff - Vertex buffer to use
 * indices - An array of indices into the vertex buffer
 * num_vtx - Number of indices from the indices array you want to draw
 * type - Primitive type to draw
 *
 * Returns:
 *
 * Number of primitives drawn
 *
 * See Also:
 *
 * <ALLEGRO_VBUFFER>, <ALLEGRO_PRIM_TYPE>, <al_prim_draw>
 */
int al_prim_draw_indexed(ALLEGRO_VBUFFER* vbuff, ALLEGRO_BITMAP* texture, int* indices, int num_vtx, int type)
{
   ASSERT(vbuff);
   ASSERT(indices);
   ASSERT(num_vtx > 0);
   ASSERT(type > 0 && type < ALLEGRO_PRIM_NUM_TYPES);
   
   ALLEGRO_BITMAP *target = al_get_target_bitmap();
   int ret = 0;
   int flags = al_get_display_flags();
   
   /* In theory, if we ever get a camera concept for this addon, the transformation into
    * view space should occur here
    */
   
   if (target->parent) {
      /*
       * For sub-bitmaps
       */
      temp_trans((float)target->xofs, (float)target->yofs);
   }
   
   if (target->flags & ALLEGRO_MEMORY_BITMAP) {
      ret =  _al_prim_draw_indexed_soft(texture, vbuff, indices, num_vtx, type);
   } else {
      if (flags & ALLEGRO_OPENGL) {
         ret =  _al_prim_draw_indexed_opengl(texture, vbuff, indices, num_vtx, type);
      } else if (flags & ALLEGRO_DIRECT3D) {
         ret =  _al_prim_draw_indexed_directx(texture, vbuff, indices, num_vtx, type);
      }
   }
   
   if (target->parent) {
      /*
       * Move it back, if rounding errors become an issue we can do a copy-restore instead
       */
      temp_trans(-(float)target->xofs, -(float)target->yofs);
   }
   
   return ret;
}

/*
 * Section: Vertex Buffer Locking
 */

/* Function: al_prim_lock_vbuff_range
 *
 * Locks a range of the vertex buffer.
 *
 * Parameters:
 *
 * vbuff - Vertex buffer to lock
 * start, end - Range to lock
 * type - Lock type; can be read, write or both
 *
 * Returns:
 *
 * 1 if locking was successful, 0 otherwise
 *
 * See Also:
 *
 * <ALLEGRO_VBUFFER>, <ALLEGRO_VBUFFER_FLAGS>
 */
int al_prim_lock_vbuff_range(ALLEGRO_VBUFFER* vbuff, int start, int end, int type)
{
   ASSERT(vbuff);
   ASSERT(start >= end);
   
   if (vbuff->flags & ALLEGRO_VBUFFER_LOCKED)
      return 0;
   if (type & ALLEGRO_VBUFFER_WRITE && !(vbuff->flags & ALLEGRO_VBUFFER_WRITE))
      return 0;
   if (type & ALLEGRO_VBUFFER_READ && !(vbuff->flags & ALLEGRO_VBUFFER_READ))
      return 0;
      
   if (vbuff->flags & ALLEGRO_VBUFFER_SOFT) {
      return _al_prim_lock_vbuff_range_soft(vbuff, start, end, type);
   }
   return 0;
}

/* Function: al_prim_lock_vbuff
 *
 * Locks a the whole vertex buffer.
 *
 * Parameters:
 *
 * vbuff - Vertex buffer to lock
 * type - Lock type; can be read, write or both
 *
 * Returns:
 *
 * 1 if locking was successful, 0 otherwise
 *
 * See Also:
 *
 * <ALLEGRO_VBUFFER>, <ALLEGRO_VBUFFER_FLAGS>
 */
int al_prim_lock_vbuff(ALLEGRO_VBUFFER* vbuff, int type)
{
   ASSERT(vbuff);
   
   return al_prim_lock_vbuff_range(vbuff, 0, vbuff->len, type);
}

/* Function: al_prim_unlock_vbuff
 *
 * Unlocks a the whole vertex buffer.
 *
 * Parameters:
 *
 * vbuff - Vertex buffer to unlock
 *
 * See Also:
 *
 * <ALLEGRO_VBUFFER>
 */
void al_prim_unlock_vbuff(ALLEGRO_VBUFFER* vbuff)
{
   ASSERT(vbuff);
   
   if (!(vbuff->flags & ALLEGRO_VBUFFER_LOCKED)) {
      return;
   }
   if (vbuff->flags & ALLEGRO_VBUFFER_SOFT) {
      _al_prim_unlock_vbuff_soft(vbuff);
   }
}

/* Function: al_prim_is_vbuff_locked
 *
 * Checks to see if the vertex buffer is locked or not.
 *
 * Parameters:
 *
 * vbuff - Vertex buffer to check
 *
 * Returns:
 *
 * 1 if it is locked, 0 otherwise
 *
 * See Also:
 *
 * <ALLEGRO_VBUFFER>
 */
int al_prim_is_vbuff_locked(ALLEGRO_VBUFFER* vbuff)
{
   ASSERT(vbuff);
   
   return vbuff->flags & ALLEGRO_VBUFFER_LOCKED;
}

/* Function: al_prim_is_vbuff_range_locked
 *
 * Checks to see if a range in the vertex buffer is locked or not.
 *
 * Parameters:
 *
 * vbuff - Vertex buffer to check
 * start, end - Range to check
 *
 * Returns:
 *
 * 0 if it not locked at all, or the range does not match, 1 otherwise
 *
 * See Also:
 *
 * <ALLEGRO_VBUFFER>
 */
int al_prim_is_vbuff_range_locked(ALLEGRO_VBUFFER* vbuff, int start, int end)
{
   ASSERT(vbuff);
   
   if (!vbuff->flags & ALLEGRO_VBUFFER_LOCKED)
      return 0;
   if (start >= vbuff->lock_start && end <= vbuff->lock_end) {
      return 1;
   }
   return 0;
}

int _al_prim_check_lock_pos(ALLEGRO_VBUFFER* vbuff, int idx, int type)
{
   ASSERT(vbuff);
   ASSERT(idx > 0 && idx < vbuff->len);
   
   if (vbuff->flags & ALLEGRO_VBUFFER_LOCKED) {
      return 0;
   } else {
      al_prim_lock_vbuff_range(vbuff, idx, idx, type);
      return 1;
   }
}

/*
 * Section: Vertex Buffer Access
 */

/* Function: al_prim_set_pos
 *
 * Sets the position variables of a vertex in a vertex buffer.
 *
 * Parameters:
 *
 * vbuff - Vertex buffer to modify
 * idx - Index of the vertex to modify
 * x, y, z - New position
 */
void al_prim_set_pos(ALLEGRO_VBUFFER* vbuff, int idx, float x, float y, float z)
{
   ASSERT(vbuff);
   
   int unlock = _al_prim_check_lock_pos(vbuff, idx, ALLEGRO_VBUFFER_WRITE);
   if (vbuff->flags & ALLEGRO_VBUFFER_SOFT) {
      _al_prim_set_pos_soft(vbuff, idx, x, y, z);
   } else {
      ASSERT(0);
   }
   if (unlock) {
      al_prim_unlock_vbuff(vbuff);
   }
}

/* Function: al_prim_set_normal
 *
 * Sets the normal variables of a vertex in a vertex buffer.
 *
 * Parameters:
 *
 * vbuff - Vertex buffer to modify
 * idx - Index of the vertex to modify
 * nx, ny, nz - New normal
 */
void al_prim_set_normal(ALLEGRO_VBUFFER* vbuff, int idx, float nx, float ny, float nz)
{
   ASSERT(vbuff);
   
   int unlock = _al_prim_check_lock_pos(vbuff, idx, ALLEGRO_VBUFFER_WRITE);
   if (vbuff->flags & ALLEGRO_VBUFFER_SOFT) {
      _al_prim_set_normal_soft(vbuff, idx, nx, ny, nz);
   } else {
      ASSERT(0);
   }
   if (unlock) {
      al_prim_unlock_vbuff(vbuff);
   }
}

/* Function: al_prim_set_uv
 *
 * Sets the uv coordinates of a vertex in a vertex buffer.
 *
 * Parameters:
 *
 * vbuff - Vertex buffer to modify
 * idx - Index of the vertex to modify
 * u,v - New texture coordinates
 */
void al_prim_set_uv(ALLEGRO_VBUFFER* vbuff, int idx, float u, float v)
{
   ASSERT(vbuff);
   
   int unlock = _al_prim_check_lock_pos(vbuff, idx, ALLEGRO_VBUFFER_WRITE);
   if (vbuff->flags & ALLEGRO_VBUFFER_SOFT) {
      _al_prim_set_uv_soft(vbuff, idx, u, v);
   } else {
      ASSERT(0);
   }
   if (unlock) {
      al_prim_unlock_vbuff(vbuff);
   }
}

/* Function: al_prim_set_vertex
 *
 * Sets a vertex in a vertex buffer.
 *
 * Parameters:
 *
 * vbuff - Vertex buffer to modify
 * idx - Index of the vertex to modify
 * vtx - New vertex definition
 */
void al_prim_set_vertex(ALLEGRO_VBUFFER* vbuff, int idx, const ALLEGRO_VERTEX *vtx)
{
   ASSERT(vbuff);
   ASSERT(vtx);
   
   int unlock = _al_prim_check_lock_pos(vbuff, idx, ALLEGRO_VBUFFER_READ);
   if (vbuff->flags & ALLEGRO_VBUFFER_SOFT) {
      _al_prim_set_vertex_soft(vbuff, idx, vtx);
   } else {
      ASSERT(0);
   }
   if (unlock) {
      al_prim_unlock_vbuff(vbuff);
   }
}

/* Function: al_prim_set_color
 *
 * Sets the colour of a vertex in a vertex buffer.
 *
 * Parameters:
 *
 * vbuff - Vertex buffer to modify
 * idx - Index of the vertex to modify
 * col - New colour
 */
void al_prim_set_color(ALLEGRO_VBUFFER* vbuff, int idx, const ALLEGRO_COLOR col)
{
   ASSERT(vbuff);
   
   int unlock = _al_prim_check_lock_pos(vbuff, idx, ALLEGRO_VBUFFER_WRITE);
   if (vbuff->flags & ALLEGRO_VBUFFER_SOFT) {
      _al_prim_set_color_soft(vbuff, idx, col);
   } else {
      ASSERT(0);
   }
   if (unlock) {
      al_prim_unlock_vbuff(vbuff);
   }
}

/* Function: al_prim_get_vertex
 *
 * Gets a vertex from the vertex buffer.
 *
 * Parameters:
 *
 * vbuff - Vertex buffer to modify
 * idx - Index of the vertex to modify
 * vtx - Vertex structure to fill out definition
 */
void al_prim_get_vertex(ALLEGRO_VBUFFER* vbuff, int idx, ALLEGRO_VERTEX *vtx)
{
   ASSERT(vbuff);
   ASSERT(vtx);
   
   int unlock = _al_prim_check_lock_pos(vbuff, idx, ALLEGRO_VBUFFER_READ);
   if (vbuff->flags & ALLEGRO_VBUFFER_SOFT) {
      _al_prim_get_vertex_soft(vbuff, idx, vtx);
   } else {
      ASSERT(0);
   }
   if (unlock) {
      al_prim_unlock_vbuff(vbuff);
   }
}

/* Function: al_prim_vbuff_flags
 *
 * Return the flags of a vertex buffer
 *
 * Parameters:
 *
 * vbuff - Vertex buffer to get flags from
 *
 * Returns:
 *
 * The flags of the buffer
 *
 * See Also:
 *
 * <ALLEGRO_VBUFFER_FLAGS>
 */
int al_prim_vbuff_flags(ALLEGRO_VBUFFER* vbuff);

/* Function: al_prim_vbuff_len
*
* Return the length of a vertex buffer
*
* Parameters:
*
* vbuff - Vertex buffer to get length of
*
* Returns:
*
* The length of the buffer
*/
int al_prim_vbuff_len(ALLEGRO_VBUFFER* vbuff);

/*
 * Section: Vertex Buffer Creation/Destruction
 */

/* Function: al_create_vbuff
 *
 * Creates a vertex buffer.
 *
 * Parameters:
 *
 * len - Number of vertices to put in this buffer
 * type -Type of the vertex buffer to create
 *
 * Returns:
 *
 * A newly created vertex buffer, 0 on failure.
 *
 * See Also:
 *
 * <ALLEGRO_VBUFFER_FLAGS>
 */
ALLEGRO_VBUFFER* al_create_vbuff(int len, int type)
{
   ALLEGRO_VBUFFER* ret = (ALLEGRO_VBUFFER*)malloc(sizeof(ALLEGRO_VBUFFER));
   ret->len = len;
   ret->flags = 0;
   if (type & ALLEGRO_VBUFFER_SOFT) {
      ret->flags |= ALLEGRO_VBUFFER_SOFT;
   } else if (type & ALLEGRO_VBUFFER_VIDEO) {
      ret->flags |= ALLEGRO_VBUFFER_VIDEO;
   }
   ret->flags |= type & ALLEGRO_VBUFFER_WRITE;
   ret->flags |= type & ALLEGRO_VBUFFER_READ;
   
   if (ret->flags & ALLEGRO_VBUFFER_SOFT) {
      _al_create_vbuff_soft(ret);
   } else {
      free(ret);
      ret = 0;
   }
   return ret;
}

/* Function: al_destroy_vbuff
 *
 * Destroys a vertex buffer.
 *
 * Parameters:
 *
 * vbuff - Vertex buffer to destroy. 0 is ignored.
 */
void al_destroy_vbuff(ALLEGRO_VBUFFER* vbuff)
{
   if (!vbuff)
      return;
   al_prim_unlock_vbuff(vbuff);
   if (vbuff->flags & ALLEGRO_VBUFFER_SOFT) {
      _al_destroy_vbuff_soft(vbuff);
   }
   free(vbuff);
}


/*
 * Section: Transfomations
 */

/* Function: al_trans_copy
 *
 * Makes a copy of a transformation
 *
 * Parameters:
 *
 * src - Source transformation
 * dest - Destination transformation
 */
void al_trans_copy(ALLEGRO_TRANSFORM* src, ALLEGRO_TRANSFORM* dest)
{
   ASSERT(src);
   ASSERT(dest);
   
   memcpy(dest, src, sizeof(ALLEGRO_TRANSFORM));
}

/* Function: al_trans_use_trans
 *
 * Sets the transformation to be used for the primitive drawing operations
 *
 * Parameters:
 *
 * trans - Transformation to use
 */
void al_trans_use_trans(ALLEGRO_TRANSFORM* trans)
{
   ASSERT(trans);
   
   al_trans_copy(trans, &_al_global_trans);
   int flags = al_get_display_flags();
   if (flags & ALLEGRO_OPENGL) {
      _al_trans_use_trans_opengl(&_al_global_trans);
   } else if (flags & ALLEGRO_DIRECT3D) {
      _al_trans_use_trans_directx(&_al_global_trans);
   }
}

/* Function: al_trans_set_identity
 *
 * Sets the transformation to be the identity transformation.
 *
 * Parameters:
 *
 * trans - Transformation to alter
 */
void al_trans_set_identity(ALLEGRO_TRANSFORM* trans);

/* Function: al_trans_build_2d
 *
 * Builds a transformation given some parameters. This call is equivalent to calling
 * the transformations in this order: scale, rotate, translate, except this method is faster.
 *
 * Parameters:
 *
 * trans - Transformation to alter
 * x, y - Translation
 * sx, sy - Scale
 * theta - Rotation angle
 */
void al_trans_build_2d(ALLEGRO_TRANSFORM* trans, float x, float y, float sx, float sy, float theta);

/* Function: al_trans_translate_2d
 *
 * Apply a translation to a transformation.
 *
 * Parameters:
 *
 * trans - Transformation to alter
 * x, y - Translation
 */
void al_trans_translate_2d(ALLEGRO_TRANSFORM* trans, float x, float y);

/* Function: al_trans_rotate_2d
 *
 * Apply a rotation to a transformation.
 *
 * Parameters:
 *
 * trans - Transformation to alter
 * theta - Rotation angle
 */
void al_trans_rotate_2d(ALLEGRO_TRANSFORM* trans, float theta);

/* Function: al_trans_rotate_2d
 *
 * Apply a scale to a transformation.
 *
 * Parameters:
 *
 * trans - Transformation to alter
 * sx, sy - Scale
 */
void al_trans_scale_2d(ALLEGRO_TRANSFORM* trans, float sx, float sy);

/* Function: al_trans_trans_vertex_2d
 *
 * Transform a vertex.
 *
 * Parameters:
 *
 * trans - Transformation to use
 * vtx - Vertex to transform
 */
void al_trans_trans_vertex_2d(ALLEGRO_TRANSFORM* trans, ALLEGRO_VERTEX* vtx);

/* Function: al_trans_trans_trans_2d
 *
 * Transform a transformation.
 *
 * Parameters:
 *
 * trans - Transformation to use
 * trans2 - Transformation to transform
 */
void al_trans_trans_trans_2d(ALLEGRO_TRANSFORM* trans, ALLEGRO_TRANSFORM* trans2);


/*
 * Section: Misc
 */

/*
 * Function: al_is_bitmap_region_locked
 */
int al_is_bitmap_region_locked(ALLEGRO_BITMAP* bmp, int x1, int y1, int w, int h)
{
   ASSERT(bmp);
   
   if (!al_is_bitmap_locked(bmp))
      return 0;
   if (x1 + w > bmp->lock_x && y1 + h > bmp->lock_y && x1 < bmp->lock_x + bmp->lock_w && y1 < bmp->lock_y + bmp->lock_h)
      return 1;
   return 0;
}

/* Function: al_prim_set_flag
 *
 * Sets a certain global flag.
 *
 * Parameters:
 *
 * flag - Flag index
 * value - New value of the flag
 *
 * See Also:
 *
 * <ALLEGRO_PRIM_FLAGS>
 */
void al_prim_set_flag(int flag, int value);

/* Function: al_prim_get_flag
 *
 * Returns the value of a certain global flag.
 *
 * Parameters:
 *
 * flag - Flag index
 *
 * Returns:
 *
 * Value of the flag
 *
 * See Also:
 *
 * <ALLEGRO_PRIM_FLAGS>
 */
int al_prim_get_flag(int flag);
